#include "Main.h"

#define CFG_ENTRIES 9

Config::Config() {
	playerMax = 0;
	backupInterval = 0;
	cfgName[0] = 0;
	cfgMotd[0] = 0;
	cfgPort[0] = 0;
	cfgPublic = true;
	cfgVerify = false;
	cfgPhysWater = false;
	cfgPhysLava = false;
	userList = NULL;
	bannedList = NULL;
}

Config::~Config() {
}

void Config::loadConfig() {
	Logger::add( "Parsing server.cfg...\n" );

	if( parseConfig() != 0 ) {
		Logger::add( "server.cfg is either corrupt or missing, writing standard template.\n" );
		writeConfig();
		parseConfig();
	}

	Logger::add( "Configuration loaded successfully!\n\n" );
}

int Config::getPlayerMax() {
	return this->playerMax;
}

int Config::getBackupInterval() {
	return this->backupInterval;
}

char *Config::getName() {
	return this->cfgName;
}

char *Config::getMotd() {
	return this->cfgMotd;
}

char *Config::getPort() {
	return this->cfgPort;
}

bool Config::getVerify() {
	return this->cfgVerify;
}

bool Config::getPublic() {
	return this->cfgPublic;
}

struct User *Config::getUserList() {
	return this->userList;
}

struct Banned *Config::getBannedList() {
	return this->bannedList;
}

void Config::writeConfig() {
	FILE *cfg;

	cfg = fopen( "server.cfg", "w" );
	fputs( "# SERVERNAME\nname=Custom Server!\n", cfg );
	fputs( "# WELCOME MESSAGE\nmotd=Welcome to my server!\n", cfg );
	fputs( "# PORT\nport=25565\n", cfg );
	fputs( "# MAX PLAYERS\nmax=16\n", cfg );
	fputs( "# PUBLIC\npublic=true\n", cfg );
	fputs( "# VERIFY NAMES\nverify-names=true\n", cfg );
	fputs( "# BACKUP INTERVAL - MINUTES\nbackup=30\n", cfg );
	fputs( "# PHYSICS - WATER\nwater=false\n", cfg );
	fputs( "# PHYSICS - LAVA\nlava=false", cfg );
	fclose(cfg);
}

int Config::parseConfig() {
	int i;
	char *chr, *tmp, buf[64];
	FILE *cfg;

	cfg = fopen( "server.cfg", "r" );

	if( ! cfg )
		/* server.cfg isn't present. */
		return 1;

	for( i = 0; i < CFG_ENTRIES; ) {
		fgets( buf, sizeof(buf), cfg );

		if( buf[0] != '#' ) {
			chr = strchr( buf, '=' );

			if( ! chr ) {
				/* Improper layout of a non-comment. */
				fclose(cfg);
				return 1;
			}

			chr++;

			tmp = strchr( buf, '\n' );
			if(tmp) *tmp = 0;
			/* else We've reached the last entry. */

			if( strncmp( buf, "name", 4 ) == 0 ) {
				strncpy( this->cfgName, chr, sizeof(this->cfgName) );
				i++;
			}

			else if( strncmp( buf, "motd", 4 ) == 0 ) {
				strncpy( this->cfgMotd, chr, sizeof(this->cfgMotd) );
				i++;
			}

			else if( strncmp( buf, "port", 4 ) == 0 ) {
				strncpy( this->cfgPort, chr, sizeof(this->cfgPort) );
				i++;
			}

			else if( strncmp( buf, "max", 3 ) == 0 ) {
				this->playerMax = atoi(chr);
				i++;
			}

			else if( strncmp( buf, "public", 6 ) == 0 ) {
				this->cfgPublic = strncmp( chr, "true", 4 ) ? false : true;
				i++;
			}

			else if( strncmp( buf, "verify-names", 12 ) == 0 ) {
				this->cfgVerify = strncmp( chr, "true", 4 ) ? false : true;
				i++;
			}

			else if( strncmp( buf, "backup", 6 ) == 0 ) {
				this->backupInterval = atoi(chr);
				this->backupInterval *= 1000;		/* Seconds. */
				this->backupInterval *= 60;			/* Minutes. */
				i++;
			}

			else if( strncmp( buf, "water", 5 ) == 0 ) {
				this->cfgPhysWater = strncmp( chr, "true", 4 ) ? false : true;
				i++;
			}

			else if( strncmp( buf, "lava", 4 ) == 0 ) {
				this->cfgPhysLava = strncmp( chr, "true", 4 ) ? false : true;
				i++;
			}


		}

		if( feof(cfg) && i < CFG_ENTRIES ) {
			/* Configuration is corrupt, since i < CFG_ENTRIES and EOF has been reached. */
			fclose(cfg);
			return 1;
		}

	}

	fclose(cfg);
	parseUserList();
	parseBannedList();
	return 0;
}

void Config::parseUserList() {
	char userbuf[64], *chr;
	struct User *tmp;
	FILE *users;

	users = fopen( "users.txt", "r" );
	if( ! users ) {
		users = fopen( "users.txt", "w" );
		fputs( "# Format follows:\n", users );
		fputs( "# username:userlevel <- 255 is *max*\n", users );
		fputs( "# Example: Notch:255\n", users );
		fclose(users);
		this->userList = NULL;
		return;
	}

	this->userList = (struct User*) malloc( sizeof(struct User) );
	tmp = this->userList;
	memset( tmp, 0, sizeof(struct User) );

	while( ! feof(users) ) {
		memset( userbuf, 0, sizeof(userbuf) );
		fgets( userbuf, sizeof(userbuf), users );
		if( userbuf[0] == '#' )
			continue;
		chr = strchr( userbuf, ':' );
		if(chr) {
			*chr = 0;
			toLowerCase(userbuf);
			strcpy( tmp->name, userbuf );
			chr++;
			tmp->level = atoi(chr) & 0x000000FF;
			tmp->next = (struct User*) malloc( sizeof(struct User) );
			tmp = tmp->next;
			memset( tmp, 0, sizeof(struct User) );
		}
	}

	fclose(users);

	tmp = this->userList;
	while(tmp) {
		if(tmp->next) {
			if( ! tmp->next->name[0] ) {
				free(tmp->next);
				tmp->next = NULL;
			}
		}
		tmp = tmp->next;
	}
}

void Config::parseBannedList() {
	char bannedbuf[64], *chr;
	struct Banned *tmp;
	FILE *banned;

	banned = fopen( "banned.txt", "r" );
	if( ! banned ) {
		banned = fopen( "banned.txt", "w" );
		fputs( "# Format follows:\n", banned );
		fputs( "# username:ip\n", banned );
		fputs( "# Remove IP, or set to zero for nameban only.\n", banned );
		fclose(banned);
		this->bannedList = NULL;
		return;
	}

	this->bannedList = (struct Banned*) malloc( sizeof(struct Banned) );
	tmp = this->bannedList;
	memset( tmp, 0, sizeof(struct Banned) );

	while( ! feof(banned) ) {
		memset( bannedbuf, 0, sizeof(bannedbuf) );
		fgets( bannedbuf, sizeof(bannedbuf), banned );
		if( bannedbuf[0] == '#' )
			continue;
		chr = strchr( bannedbuf, ':' );
		if(chr) {
			*chr = 0;
			toLowerCase(bannedbuf);
			strcpy( tmp->name, bannedbuf );
			chr++;
			if(*chr)
				strcpy( tmp->ip, chr );
			tmp->next = (struct Banned*) malloc( sizeof(struct Banned) );
			tmp = tmp->next;
			memset( tmp, 0, sizeof(struct Banned) );
		}
	}

	fclose(banned);

	tmp = this->bannedList;
	while(tmp) {
		if(tmp->next) {
			if( ! tmp->next->name[0] ) {
				free(tmp->next);
				tmp->next = NULL;
			}
		}
		tmp = tmp->next;
	}
}
