#include "Main.h"

Level::Level() {
}

Level::~Level() {
	free(this->serverLevel);
}

/* Will load up gzipped map fileName */
/* and store in unsigned char *serverLevel. */
void Level::loadMap( char *fileName ) {
	unsigned char *compressedMap;
	FILE *map_in;

	Logger::add( "Loading map...\n" );

	map_in = fopen( fileName, "rb" );

	if( ! map_in ) {
	    Logger::add( "server_level.dat not found, generating a default flatmap.\n" );
		newMap( 128, 128, 128 );
	}

	else {
		fseek( map_in, -4, SEEK_END );
		fread( &this->decompressed_mapLen, 1, 4, map_in );
		this->compressed_mapLen = ftell(map_in);
		rewind(map_in);

		compressedMap = (unsigned char*) malloc(this->compressed_mapLen);
		fread( compressedMap, 1, this->compressed_mapLen, map_in );
		fclose(map_in);

		this->serverLevel = (unsigned char*) malloc(this->decompressed_mapLen);
		this->decompressed_mapLen = gzip_decompress( this->serverLevel, compressedMap, this->compressed_mapLen );
		free(compressedMap);
		fixMap();
	}
}

/* Will "reformat" the old map, removing a large header */
/* and a smaller trailer. Will save us about 2 kB on large */
/* maps and even more when they are compressed, so saveMap */
/* wont have to work so hard. Also adds our mini-header */
/* containing 3 unsigned chars (identification), and another 13 unsigned chars */
/* for spawn/dimensions XYZ unsigned shorts */
void Level::fixMap() {

	if( ! identifyMap() ) {

		if( ! deserialize( this->serverLevel, this->decompressed_mapLen ) ) {
			Logger::add( "Map is in a completely unknown format. We're sorry!\n\n" );
			exit(0);
		}

		removeHeader();
		addHeader();
		Logger::add( "Success! Corrected map WILL overwrite server_level.dat\n" );
	}

	else {
		memcpy( &this->bDimX, this->serverLevel+3, 2 );
		memcpy( &this->bDimY, this->serverLevel+5, 2 );
		memcpy( &this->bDimZ, this->serverLevel+7, 2 );
		memcpy( &this->spawnX, this->serverLevel+9, 2 );
		memcpy( &this->spawnY, this->serverLevel+11, 2 );
		memcpy( &this->spawnZ, this->serverLevel+13, 2 );
		this->spawnRotX = *(this->serverLevel+15);
		this->spawnRotY = *(this->serverLevel+16);
		this->lDimX = ntohs(this->bDimX);
		this->lDimY = ntohs(this->bDimY);
		this->lDimZ = ntohs(this->bDimZ);
	}

	Logger::add( "Map loaded, have fun!\n\n" );
}

bool Level::identifyMap() {
	unsigned char ourHeader[] = { 0xBE, 0xEE, 0xEF };

	if( memcmp( this->serverLevel, ourHeader, sizeof(ourHeader) ) )
		return false;

	return true;
}

void Level::removeHeader() {
	unsigned char serializedEnd[] = { 0x00, 0x78, 0x70 };
	unsigned char *ptr, *tmp;

	ptr = memstr( this->serverLevel, this->decompressed_mapLen, serializedEnd, 3 );
	tmp = ptr;

	while(ptr) {
		ptr += 3;
		tmp = ptr;
		ptr = memstr( ptr, this->decompressed_mapLen - (ptr - this->serverLevel), serializedEnd, 3 );
	}

	this->decompressed_mapLen -= ( tmp - this->serverLevel );
	this->decompressed_mapLen += OUR_HEADERLEN;
	memmove(this->serverLevel, tmp - OUR_HEADERLEN, this->decompressed_mapLen);
	this->decompressed_mapLen = (OUR_HEADERLEN + 4) + (this->lDimX * this->lDimY * this->lDimZ);
	realloc(this->serverLevel, this->decompressed_mapLen);
}

void Level::addHeader() {
	unsigned char ourHeader[] = { 0xBE, 0xEE, 0xEF };

	memcpy( this->serverLevel, ourHeader, 3 );
	memcpy( this->serverLevel+3, &this->bDimX, 2 );
	memcpy( this->serverLevel+5, &this->bDimY, 2 );
	memcpy( this->serverLevel+7, &this->bDimZ, 2 );
	memcpy( this->serverLevel+9, &this->spawnX, 2 );
	memcpy( this->serverLevel+11, &this->spawnY, 2 );
	memcpy( this->serverLevel+13, &this->spawnZ, 2 );
	*(this->serverLevel+15) = this->spawnRotX;
	*(this->serverLevel+16) = this->spawnRotY;
}

bool Level::setSpawn( unsigned short x, unsigned short y, unsigned short z, unsigned char rotx, unsigned char roty ) {
	
	/* If invalid position return false. */
	
	this->spawnX = x;
	this->spawnY = y;
	this->spawnZ = z;
	this->spawnRotX = rotx;
	this->spawnRotY = roty;

	memcpy( this->serverLevel+9, &this->spawnX, 2 );
	memcpy( this->serverLevel+11, &this->spawnY, 2 );
	memcpy( this->serverLevel+13, &this->spawnZ, 2 );
	*(this->serverLevel+15) = this->spawnRotX;
	*(this->serverLevel+16) = this->spawnRotY;

	return true;
}

void Level::saveMap( char *fileName ) {
	unsigned char *compressedMap;
	FILE *map_out;

	compressedMap = (unsigned char *) malloc( this->compressed_mapLen + 16384 );
	this->compressed_mapLen = gzip_compress( compressedMap, this->serverLevel, this->decompressed_mapLen );

	map_out = fopen( fileName, "wb" );
	fwrite( compressedMap, 1, this->compressed_mapLen, map_out );
	fclose(map_out);
	free(compressedMap);
}

bool Level::deserialize( unsigned char *Object, unsigned long objectLen ) {
	unsigned long i, offset;
	unsigned long dx = 0, dy = 0, dz = 0;
	unsigned long sx = 0, sy = 0, sz = 0;
	unsigned short tmp;

	unsigned char serializedStart[] = { 0xAC, 0xED };
	unsigned char serializedEnd[] = { 0x78, 0x70 };
	unsigned char *temp;

	for( i = 0; i < objectLen; i++ ) {

		if( memcmp( Object+i, serializedStart, 2 ) == 0 ) {

			Logger::add( "Map contains Java serialization.\n" );
			Logger::add( "Will attempt to deserialize, might result in crashes, and worse.\n" );

			/* Jump past Java-gaystuff and hope that it's constant in each map. */
			/* Or just jump by, and find out what purpose it fills later! */

			Object += i;
			Object += 6;
			memcpy( &tmp, Object, 2 );
			Object += htons(tmp);
			Object += 13;

			/* We're certain that we're iterating over serialized memory. */
			/* That means there has to be a termination: serializedEnd. */
			/* Find it, to use as offset for each variable we extract. */

			for( temp = Object; ; temp++ ) {
				if( memcmp( temp, serializedEnd, 2 ) == 0 ) {
					temp += 2;
					break;
				}
			}

			offset = 0;

			for( ;; ) {

				/* This is needed because the first variable entry - */
				/* Isn't supposed to add on offset... */

				if( *Object == 'Z' )
					offset++;

				else if( *Object == 'I' || *Object == 'F' )
					offset += 4;

				else if( *Object == 'J' )
					offset += 8;

				Object++;
				memcpy( &tmp, Object, 2 );
				Object += 2;

				if( memcmp( Object, "width", 5 ) == 0 )
					memcpy( &dx, temp+offset-4, 4 );

				else if( memcmp(Object, "depth", 5 ) == 0 )
					memcpy( &dy, temp+offset-4, 4 );

				else if( memcmp(Object, "height", 6 ) == 0 )
					memcpy( &dz, temp+offset-4, 4 );

				else if( memcmp(Object, "xSpawn", 6 ) == 0 )
					memcpy( &sx, temp+offset-4, 4 );

				else if( memcmp( Object, "ySpawn", 6 ) == 0 )
					memcpy( &sy, temp+offset-4, 4 );

				else if( memcmp( Object, "zSpawn", 6 ) == 0 ) {
					memcpy( &sz, temp+offset-4, 4 );
					break;
				}

				Object += htons(tmp);
			}

			dx = ntohl(dx);
			dy = ntohl(dy);
			dz = ntohl(dz);
			sx = ntohl(sx);
			sy = ntohl(sy);
			sz = ntohl(sz);

			this->lDimX = (unsigned short) dx;
			this->lDimY = (unsigned short) dy;
			this->lDimZ = (unsigned short) dz;
			this->spawnX = (unsigned short) sx;
			this->spawnY = (unsigned short) sy;
			this->spawnZ = (unsigned short) sz;

			this->spawnX *= 32;
			this->spawnY *= 32;
			this->spawnZ *= 32;
			this->spawnX += 16;
			this->spawnZ += 16;

			this->bDimX = htons(this->lDimX);
			this->bDimY = htons(this->lDimY);
			this->bDimZ = htons(this->lDimZ);
			this->spawnX = htons(this->spawnX);
			this->spawnY = htons(this->spawnY);
			this->spawnZ = htons(this->spawnZ);

			return true;
		}
	}

	return false;
}

/* Creates a new flat map */
void Level::newMap(unsigned short sizeX, unsigned short sizeY, unsigned short sizeZ) {
	unsigned long size;
	unsigned short i, j, k;

	while(sizeX % 8)
		sizeX++;

	while(sizeY % 8)
		sizeY++;

	while(sizeZ % 8)
		sizeZ++;

	size = sizeX * sizeY * sizeZ;

	this->decompressed_mapLen = (size + OUR_HEADERLEN + 4);
	this->compressed_mapLen = this->decompressed_mapLen;

	this->serverLevel = (unsigned char*) malloc(this->decompressed_mapLen);

	if( ! this->serverLevel ) {
		Logger::add( "Couldn't allocate enough memory to store the map!\n" );
		exit(0);
	}

	memset( this->serverLevel, W_BLOCK_AIR, this->decompressed_mapLen );

	// Do this now, blockIndex() wants dimX, dimY and dimZ to be set
	this->lDimX = sizeX;
	this->lDimY = sizeY;
	this->lDimZ = sizeZ;
	this->spawnX = sizeX/2;
	this->spawnY = (sizeY+2)/2;
	this->spawnZ = sizeZ/2;
	this->spawnRotX = 0;
	this->spawnRotY = 0;

	this->spawnX *= 32;
	this->spawnY *= 32;
	this->spawnZ *= 32;
	this->spawnX += 16;
	this->spawnY += 16;

	this->bDimX = htons(this->lDimX);
	this->bDimY = htons(this->lDimY);
	this->bDimZ = htons(this->lDimZ);

	for( j = 0; j < (sizeY / 2) - 1; j++ ) {
		for( i = 0; i < sizeX; i++ ) {
			for( k = 0; k < sizeZ; k++ ) {
				serverLevel[blockIndex( i, j, k )] = W_BLOCK_DIRT;
			}
		}
	}

	for( j = ( sizeY / 2) - 1; j < (sizeY / 2); j++ ) {
		for( i = 0; i < sizeX; i++ ) {
			for( k = 0; k < sizeZ; k++ ) {
				serverLevel[blockIndex( i, j, k )] = W_BLOCK_GRASS;
			}
		}
	}

	/* Add the size (dimX * Y * Z) to the header. */
	size = htonl(size);
	memcpy( this->serverLevel+OUR_HEADERLEN, &size, 4 );

	/* This is our last endianshift. */
	this->spawnX = htons(this->spawnX);
	this->spawnY = htons(this->spawnY);
	this->spawnZ = htons(this->spawnZ);

	addHeader();
}
