#ifndef _LEVEL_H
#define _LEVEL_H

#include "defines.h"

/* BLOCK TYPES */
#define W_BLOCK_AIR		 0x00
#define W_BLOCK_STONE	 0x01
#define W_BLOCK_GRASS	 0x02
#define W_BLOCK_DIRT	 0x03
#define W_BLOCK_COBBLE	 0x04
#define W_BLOCK_WOOD	 0x05
#define W_BLOCK_PLANT	 0x06
#define W_BLOCK_ROCK	 0x07
#define W_BLOCK_WATER	 0x08
#define W_BLOCK_SWATER	 0x09
#define W_BLOCK_LAVA	 0x0A
#define W_BLOCK_SLAVA	 0x0B
#define W_BLOCK_SAND	 0x0C
#define W_BLOCK_GRAVEL	 0x0D
#define W_BLOCK_GOLDORE	 0x0E
#define W_BLOCK_IRONORE	 0x0F
#define W_BLOCK_COAL	 0x10
#define W_BLOCK_TREE	 0x11
#define W_BLOCK_LEAVES	 0x12
#define W_BLOCK_SPONGE	 0x13
#define W_BLOCK_GLASS	 0x14
#define W_BLOCK_REDC	 0x15
#define W_BLOCK_ORANGEC	 0x16
#define W_BLOCK_YELLOWC	 0x17
#define W_BLOCK_LGREENC	 0x18
#define W_BLOCK_GREENC	 0x19
#define W_BLOCK_AGREENC	 0x1A
#define W_BLOCK_CYANC	 0x1B
#define W_BLOCK_BLUEC	 0x1C
#define W_BLOCK_PURPLEC	 0x1D
#define W_BLOCK_INDIGOC	 0x1E
#define W_BLOCK_VIOLETC  0x1F
#define W_BLOCK_MAGENTAC 0x20
#define W_BLOCK_PINKC	 0x21
#define W_BLOCK_BLACKC	 0x22
#define W_BLOCK_GRAYC	 0x23
#define W_BLOCK_WHITEC	 0x24
#define W_BLOCK_YFLOWER	 0x25
#define W_BLOCK_RFLOWER	 0x26
#define W_BLOCK_BSHROOM	 0x27
#define W_BLOCK_RSHROOM	 0x28
#define W_BLOCK_GOLD	 0x29
#define W_BLOCK_IRON	 0x2A
#define W_BLOCK_DOUBSTAIR 0x2B
#define W_BLOCK_STAIR	 0x2C
#define W_BLOCK_BRICK	 0x2D
#define W_BLOCK_TNT		 0x2E
#define W_BLOCK_BOOKS	 0x2F
#define W_BLOCK_MOSSCOB	 0x30
#define W_BLOCK_OBSIDIAN 0x31

class Level {

public:

	Level();
	~Level();

	void loadMap( char *fileName );
	void saveMap( char *fileName );
	bool setSpawn( unsigned short x, unsigned short y, unsigned short z, unsigned char rotx, unsigned char roty );

	/* Inlines for speed. */

	inline bool validBlock(unsigned short x, unsigned short y, unsigned short z) {
		if( x > this->lDimX )
			return false;
		if( y >= this->lDimY )
			return false;
		if( z > this->lDimZ )
			return false;

		return true;
	}

	inline unsigned long blockIndex(unsigned short x, unsigned short y, unsigned short z) {
		return ( (y * this->lDimZ + z) * this->lDimX + x ) + OUR_HEADERLEN + 4;
	}

	inline void setBlock(unsigned long offset, unsigned char type) {
		this->serverLevel[offset] = type;
	}

	inline void setBlock(unsigned short x, unsigned short y, unsigned short z, unsigned char type) {
		this->serverLevel[blockIndex( x, y, z )] = type;
	}

	inline unsigned char getBlock(unsigned short x, unsigned short y, unsigned short z ) {
		return this->serverLevel[blockIndex( x, y, z )];
	}

	inline unsigned long getCompressedMapLen() {
		return this->compressed_mapLen;
	}

	inline void setCompressedMapLen( unsigned long newLen ) {
		this->compressed_mapLen = newLen;
	}

	inline unsigned long getDecompressedMapLen() {
		return this->decompressed_mapLen;
	}

	inline unsigned short getDimX() {
		return this->bDimX;
	}

	inline unsigned short getDimY() {
		return this->bDimY;
	}
	
	inline unsigned short getDimZ() {
		return this->bDimZ;
	}
	
	inline unsigned short getSpawnX() {
		return this->spawnX;
	}
	
	inline unsigned short getSpawnY() {
		return this->spawnY;
	}
	
	inline unsigned short getSpawnZ() {
		return this->spawnZ;
	}

	inline unsigned char getRotX() {
		return this->spawnRotX;
	}

	inline unsigned char getRotY() {
		return this->spawnRotY;
	}

	inline unsigned char *getServerLevel() {
		return this->serverLevel;
	}

private:

	void fixMap();
	bool identifyMap();
	bool deserialize( unsigned char *Object, unsigned long objectLen );
	void removeHeader();
	void addHeader();

	void newMap( unsigned short sizeX, unsigned short sizeY, unsigned short sizeZ );

	unsigned long compressed_mapLen;
	unsigned long decompressed_mapLen;

	/* Little and big endian. */
	/* Better then doing endianshifts all the f'n time... */
	unsigned short bDimX, bDimY, bDimZ;
	unsigned short lDimX, lDimY, lDimZ;
	unsigned short spawnX, spawnY, spawnZ;
	unsigned char spawnRotX, spawnRotY;

	unsigned char *serverLevel;

};

#endif
