#include "Main.h"
#include "srvMessage.h"

Server::Server() {
	playerNum = 0;
	ticks_ping = 0;
	ticks_beat = 0;
	ticks_savemap = 0;
	ticks_backupmap = 0;
	salt = 0;
	isOperated = false;

	memset(&timeout, 0, sizeof(struct timeval));
	memset(freeIdList, 0, UCHAR_MAX);

	CreateDirectory("backups", NULL);
}

Server::~Server() {
	struct User *utmp;
	struct Banned *btmp;

	delete this->lvl;

	while(this->userList) {
		utmp = this->userList;
		this->userList = this->userList->next;
		free(utmp);
	}

	while(this->bannedList) {
		btmp = this->bannedList;
		this->bannedList = this->bannedList->next;
		free(btmp);
	}

	closesocket(sock_listener);
	freeaddrinfo(heartbeat);
	freeaddrinfo(listener);
	WSACleanup();
}

void Server::init() {
	struct addrinfo hints;
	FILE *guestarea;

	Logger::start();
	Logger::add("Initializing server!\n\n");

	/* Retrieve configuration. */
	cfg = new Config();

	cfg->loadConfig();

	playerMax = cfg->getPlayerMax();
	if(playerMax > 255)
		playerMax = 255;

	backupInterval = cfg->getBackupInterval();

	strcpy(srvName, cfg->getName());
	strcpy(srvMotd, cfg->getMotd());
	strcpy(srvPort, cfg->getPort());

	srvPublic = cfg->getPublic();
	srvVerify = cfg->getVerify();
	userList = cfg->getUserList();
	bannedList = cfg->getBannedList();

	delete cfg;
	/* ! Config */

	/* Guestarea */
	guestarea = fopen("guest.dat", "rb");

	if(guestarea == NULL) {
		memset(guestFrom, 0, sizeof(unsigned short) * 3);
		memset(guestTo, 0, sizeof(unsigned short) * 3);
		srvGuest = false;
	}

	else {
		fread(guestFrom, 2, 3, guestarea);
		fread(guestTo, 2, 3, guestarea);
		srvGuest = true;
		fclose(guestarea);
	}
	/* ! Guestarea */

	if(srvVerify == true) {
		srand(GetTickCount());
		salt = GetTickCount() - rand();
	}

	lvl = new Level();
	lvl->loadMap("server_level.dat");

	/* Initialize netcode... */
	WSAStartup(MAKEWORD(2, 2), &wsaData);
	heartbeat = (struct addrinfo *) malloc(sizeof(struct addrinfo));
	listener = (struct addrinfo *) malloc(sizeof(struct addrinfo));
	memset(&hints, 0, sizeof(struct addrinfo));

	/* HEARTBEAT ADDRINFO */
	hints.ai_family = PF_INET;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = IPPROTO_TCP;
	getaddrinfo("www.minecraft.net", "80", &hints, &heartbeat);

	/* LISTENER ADDRINFO */
	hints.ai_flags = AI_PASSIVE;
	getaddrinfo(NULL, srvPort, &hints, &listener);

	/* Set up and bind a listener, add to fd_set master. */
	sock_listener = socket(listener->ai_family, listener->ai_socktype, listener->ai_protocol);
	bind(sock_listener, listener->ai_addr, listener->ai_addrlen);
	listen(sock_listener, 1);
	FD_ZERO(&master);
	FD_SET(sock_listener, &master);
}

bool Server::step() {
	unsigned int i;
	int j;
	char buffer[PACKLEN_SRVINFO];

	readfds = master;
	select(0, &readfds, 0, 0, &timeout);

	/* Check how our players are doing */
	for(i = 0; i < playerList.size(); i++) {
		if(FD_ISSET(playerList.at(i).sock, &readfds)) {

			j = recv(playerList.at(i).sock, buffer, 1, MSG_PEEK);

			if( j == 0 || j == SOCKET_ERROR ) {
				memset(buffer, 0x20, PACKLEN_CHAT);
				buffer[0] = 0x0D;
				buffer[1] = 0x00;
				memcpy(buffer+2, "&e", 2);
				memcpy(buffer+4, playerList.at(i).playerName, playerList.at(i).namelen);
				memcpy(buffer+4+playerList.at(i).namelen, " left the game!", 15);

				if(playerList.at(i).hidden == true)
					broadcastLevelSelective(buffer, PACKLEN_CHAT, &playerList.at(i), false);
				else
					broadcast(buffer, PACKLEN_CHAT, NULL);

				releasePlayer(&playerList.at(i));
			}

			else {
				if(*buffer == 0x00)
					j = PACKLEN_SRVINFO;
				else if(*buffer == 0x05)
					j = PACKLEN_BLOCK;
				else if(*buffer == 0x08)
					j = PACKLEN_MOVE;
				else if(*buffer == 0x0D)
					j = PACKLEN_CHAT;
				else continue;

				recv(playerList.at(i).sock, buffer, j, 0);
				parsePacket(&playerList.at(i), buffer);
			}
		}
	}

	ticks = GetTickCount();
	if(ticks > ticks_ping) {
		broadcast("\x01", 1, NULL);
		ticks_ping = ticks + 500;
	}

	/* Do we need this kind of precision or ... */
	/* ... could we cut down all those calls? */
	/* ticks = GetTickCount(); */
	if(ticks > ticks_beat) {
		beat();
		ticks_beat = ticks + 45000;
	}

	/* ticks = GetTickCount(); */
	if(ticks > ticks_savemap ) {
		lvl->saveMap("server_level.dat");
		ticks_savemap = ticks + 60000;
		memset(buffer, 0, sizeof(buffer));
		sprintf(buffer, "Map saved! Players: %d/%d\n", this->playerNum, this->playerMax);
		Logger::add(buffer);
	}
	
	if(ticks > ticks_backupmap) {

		if(this->playerNum > 0) {
			char current_mapname[40];
			SYSTEMTIME ts;

			GetLocalTime(&ts);
			memset(current_mapname, 0, sizeof(current_mapname));
			sprintf(current_mapname, "backups\\backup-%d_%d_%d_%d", ts.wMonth, ts.wDay, ts.wHour, ts.wMinute);
			lvl->saveMap(current_mapname);
			Logger::add("Map backup: Saved.\n");
		}

		else 
			Logger::add("Map backup: No players online, skipping.\n"); 
        
		ticks_backupmap = ticks + (this->backupInterval > 0 ? this->backupInterval : 1800000);
	}       

	/* Accept new connections */
	if(FD_ISSET(sock_listener, &readfds)) {
	    Player newPlayer;
		memset(&newPlayer, 0, sizeof(Player));

		i = sizeof(newPlayer.clientinfo);
		newPlayer.sock = accept(sock_listener, (struct sockaddr*) &newPlayer.clientinfo, (int*) &i);

		if(this->playerNum >= this->playerMax)
			kickPlayer(&newPlayer, "Server is full!", false);

		else {
			playerList.push_back(newPlayer);
			FD_SET(playerList.back().sock, &master);

			playerList.back().id_byte = getFreeId();
			playerList.back().id_privsay = playerList.back().id_byte;
			playerList.back().posx = lvl->getSpawnX();
			playerList.back().posy = lvl->getSpawnY();
			playerList.back().posz = lvl->getSpawnZ();
			playerList.back().rotx = lvl->getRotX();
			playerList.back().roty = lvl->getRotY();
			this->playerNum = playerList.size();
		}
	}

	Sleep(10);
	return true;
}

bool Server::drawBlocks(char block, unsigned short *drawFrom, unsigned short *drawTo, unsigned char drawMax) {
	unsigned short tmp, tempFrom[3], tempTo[3], draw[3];
	unsigned char count = 0;
	char drawbuf[8];

	if(drawFrom[0] < drawTo[0]) {
		tempFrom[0] = drawFrom[0];
		tempTo[0] = drawTo[0];
	}
		
	else {
		tempFrom[0] = drawTo[0];
		tempTo[0] = drawFrom[0];
	}

	if(drawFrom[1] < drawTo[1]) {
		tempFrom[1] = drawFrom[1];
		tempTo[1] = drawTo[1];
	}
		
	else { 
		tempFrom[1] = drawTo[1];
		tempTo[1] = drawFrom[1];
	}

	if(drawFrom[2] < drawTo[2]) {
		tempFrom[2] = drawFrom[2];
		tempTo[2] = drawTo[2];
	}
		
	else {
		tempFrom[2] = drawTo[2];
		tempTo[2] = drawFrom[2];
	}

	drawbuf[0] = 0x06;
	drawbuf[7] = block;

	for( draw[0] = tempFrom[0]; draw[0] <= tempTo[0]; draw[0]++ ) {
		for( draw[1] = tempFrom[1]; draw[1] <= tempTo[1]; draw[1]++ ) {
			for( draw[2] = tempFrom[2]; draw[2] <= tempTo[2]; draw[2]++ ) {

				if(lvl->getBlock(draw[0], draw[1], draw[2]) == block) {
					count++;
					continue;
				}

				if(drawMax != 0 && count > drawMax)
					return false;

				tmp = htons(draw[0]);
				memcpy( drawbuf+1, &tmp, 2 );

				tmp = htons(draw[1]);
				memcpy( drawbuf+3, &tmp, 2 );

				tmp = htons(draw[2]);
				memcpy( drawbuf+5, &tmp, 2 );

				broadcast(drawbuf, 8, NULL);
				lvl->setBlock(draw[0], draw[1], draw[2], block);

				count++;
			}
		}
	}

	return true;
}

char* Server::levelToClass(unsigned char userlevel) {
	if( userlevel > 0) {
		userlevel = (userlevel-1)/50;
		switch(userlevel) {
			case 0:
				return " &fUser";
			case 1:
				return " &aTrusted User";
			case 2:
				return "n &9OP";
			case 3:
				return " &9Trusted OP";
			case 4:
				return "n &cAdmin";
			case 5:
				return "n &8Elite";
		}
	}

	return " &7Guest";
}

void Server::setGuestArea() {
	FILE *guestarea;

	this->srvGuest = true;

	guestarea = fopen("guest.dat", "wb");
	fwrite(guestFrom, 2, 3, guestarea);
	fwrite(guestTo, 2, 3, guestarea);
	fclose(guestarea);
}

bool Server::isGuestBlock(unsigned short x, unsigned short y, unsigned short z) {
	if(x < this->guestFrom[0] && x < this->guestTo[0])
		return false;
	if(y < this->guestFrom[1] && y < this->guestTo[1])
		return false;
	if(z < this->guestFrom[2] && z < this->guestTo[2])
		return false;

	if(x > this->guestFrom[0] && x > this->guestTo[0])
		return false;
	if(y > this->guestFrom[1] && y > this->guestTo[1])
		return false;
	if(z > this->guestFrom[2] && z > this->guestTo[2])
		return false;

	if(this->srvGuest == false)
		return false;
	
	return true;
}

void Server::teleportPlayer( Player *toPlayer, Player *who ) {
	char tpbuf[10];

	tpbuf[0] = 0x08;
	tpbuf[1] = -1;

	memcpy( tpbuf+2, &toPlayer->posx, 2 );
	memcpy( tpbuf+4, &toPlayer->posy, 2 );
	memcpy( tpbuf+6, &toPlayer->posz, 2 );
	tpbuf[8] = toPlayer->rotx;
	tpbuf[9] = toPlayer->roty;
	send( who->sock, tpbuf, 10, 0 );
	tpbuf[1] = who->id_byte;
	broadcast( tpbuf, 10, NULL );
}

void Server::unknownCommand( Player *player ) {
	char chatbuf[PACKLEN_CHAT];
	memset(chatbuf, 0x20, PACKLEN_CHAT);
	chatbuf[0] = 0x0D;
	chatbuf[1] = 0x00;
	memcpy(chatbuf+2, srvMessage::UNKNOWN_CMD, 18);
	send(player->sock, chatbuf, PACKLEN_CHAT, 0);
}

void Server::fakeLog(Player *player,  bool out) {
	char logbuf[PACKLEN_SPAWN];

	if(out) {
		memset( logbuf, 0x20, PACKLEN_CHAT );

		logbuf[0] = 0x0D;
		logbuf[1] = 0x00;

		memcpy(logbuf+2, "&e", 2 );
		memcpy(logbuf+4, player->playerName, player->namelen);
		memcpy(logbuf+4+player->namelen, " left the game!", 15);
		broadcastLevelSelective( logbuf, PACKLEN_CHAT, player, true );

		memcpy( logbuf+4+player->namelen, " is now hidden!", 15 );
		broadcastLevelSelective( logbuf, PACKLEN_CHAT, player, false );

		logbuf[0] = 0x0C;
		logbuf[1] = player->id_byte;
		broadcastLevelSelective( logbuf, 2, player, true );
	}

	else {
		memset( logbuf, 0x20, PACKLEN_SPAWN );

		logbuf[0] = 0x07;
		logbuf[1] = player->id_byte;

		logbuf[2] = '&';
		logbuf[3] = getColorCode(player->userlevel);
		memcpy(logbuf+4, player->playerName, player->namelen);
		memcpy(logbuf+66, &player->posx, 2);
		memcpy(logbuf+68, &player->posy, 2);
		memcpy(logbuf+70, &player->posz, 2);
		logbuf[72] = player->rotx;
		logbuf[73] = player->roty;
		broadcastLevelSelective(logbuf, PACKLEN_SPAWN, player, true);

		logbuf[0] = 0x0D;
		logbuf[1] = 0x00;
		memcpy(logbuf+2, "&e", 2);
		memcpy(logbuf+4, player->playerName, player->namelen);
		memcpy(logbuf+player->namelen+4, " joined the game!", 17);
		broadcastLevelSelective(logbuf, PACKLEN_CHAT, player, true);
		memcpy(logbuf+player->namelen+4, " is no longer hidden!", 21);
		broadcastLevelSelective(logbuf, PACKLEN_CHAT, player, false);
	}

}

/* TODO: Prebuilt buffers, creating them realtime is too slow. */
bool Server::playerCommand( Player *player, char *buffer ) {
	unsigned int j;
	char chatbuf[PACKLEN_CHAT], *chr;
	char sprintfbuf[PACKLEN_CHAT];
	
	memset(sprintfbuf, 0x00, PACKLEN_CHAT);

	if( *buffer == 0x20 ) {
		unknownCommand(player);
		return false;
	}

	memset( chatbuf, 0x20, PACKLEN_CHAT );
	chatbuf[0] = 0x0D;
	chatbuf[1] = 0x00;

	chr = strchr( buffer, 0x20 );
	if(! chr) return false;
	chr++;					/* Now points to the 2nd param. */
	j = chr - buffer;		/* This gives us the length of the "playerCommand" with the / included */

	if( strncmp( buffer, "me ", 3 ) == 0 ) {

		if( *chr == 0x20 )
			return false;

		memcpy(chatbuf+2, "* &b", 4);
		memcpy(chatbuf+6, player->playerName, player->namelen);
		memcpy(chatbuf+player->namelen+7, chr, 66-7-player->namelen-j);
		broadcast(chatbuf, PACKLEN_CHAT, NULL);
		return true;
	}

	else if( strncmp( buffer, "set ", 4 ) == 0 || strncmp( buffer, "setblock ", 9 ) == 0 ) {

		player->setblock = true;

		if( strncmp( chr, "g ", 2 ) == 0 || strncmp( chr, "grass ", 6 ) == 0 )
			player->block = W_BLOCK_GRASS;

		else if( strncmp( chr, "l ", 2 ) == 0 || strncmp( chr, "lava ", 5 ) == 0 )
			player->block = W_BLOCK_LAVA;

		else if( strncmp( chr, "w ", 2 ) == 0 || strncmp( chr, "water ", 6 ) == 0 )
			player->block = W_BLOCK_WATER;
		
		else if( strncmp( chr, "a ", 2 ) == 0 || strncmp( chr, "air ", 4 ) == 0 )
			player->block = W_BLOCK_AIR;

		else player->setblock = false;

		memcpy( chatbuf+2, srvMessage::BLOCK_PLACING, 34 );

		if(player->setblock == true) {
			switch(player->block) {
				case W_BLOCK_GRASS:
					memcpy( chatbuf+36, srvMessage::BLOCK_GRASS, 6 );
					break;
				case W_BLOCK_WATER:
					memcpy( chatbuf+36, srvMessage::BLOCK_WATER, 6 );
					break;
				case W_BLOCK_LAVA:
					memcpy( chatbuf+36, srvMessage::BLOCK_LAVA, 5 );
					break;
				case W_BLOCK_AIR:
					memcpy( chatbuf+36, srvMessage::BLOCK_AIR, 4 );
					break;
			}
		}

		else
			memcpy( chatbuf+2, srvMessage::BLOCK_NORMAL, 34 );

		send( player->sock, chatbuf, PACKLEN_CHAT, 0 );
		return true;
	}

	/* Thanks a lot to you, Dandistine, for helping us with such a beautiful solution. */
	/* Your help is much needed and appreciated.                      // vLK & Unison. */
	else if( strncmp( buffer, "help ", 5 ) == 0 ) {
        unsigned int helpFrom, helpTo;

		if( *chr == 0x20 ) {
			helpFrom = 0;
			helpTo = 8;
		}

		else {
			buffer = chr;
			chr = strchr(buffer, 0x20);
			if(chr) *chr = 0;
			else return false;

			toLowerCase(buffer);

			if( strncmp(buffer, "normal", 6) == 0 ) {
				helpFrom = 9;
				helpTo = 13;
			}

			else if( strncmp(buffer, "ranks", 5) == 0 ) {
				helpFrom = 14;
				helpTo = 21;
			}

			else if( strncmp(buffer, "build", 5) == 0 ) {
				helpFrom = 22;
				helpTo = 26;
			}

			else if( strncmp(buffer, "users", 5) == 0 ) {
				helpFrom = 27;
				helpTo = 39;
			}
		
			else return false;
		}
		
		for( ; helpFrom <= helpTo; ++helpFrom )
			send( player->sock, srvMessage::HELP_MSG[helpFrom], PACKLEN_CHAT, 0 );

		return true;        
	}

	else if( strncmp( buffer, "pinfo ", 6 ) == 0 ) {
		Player *tmpPlayer;

		if( *chr == 0x20)
			return false;
		
		buffer = chr;
		chr = strchr( buffer, 0x20 );
		if(chr) *chr = 0;
		else return false;

		tmpPlayer = getPlayerByName( buffer );

		if(tmpPlayer)
            sprintf(sprintfbuf, "&e%s is a%s &e(%d)", tmpPlayer->playerName, levelToClass(tmpPlayer->userlevel), tmpPlayer->userlevel);

		else {
			User *tmpUser;
			tmpUser = getUserByName( toLowerCase(buffer) );
			if(tmpUser) {
				sprintf(sprintfbuf, "&e%s is a%s &e(%d)", tmpUser->name, levelToClass(tmpUser->level), tmpUser->level);
			}
			else
			sprintf( sprintfbuf, "&eNo information found.");
		}

		memcpy(chatbuf+2, sprintfbuf, strlen(sprintfbuf));
		send(player->sock, chatbuf, PACKLEN_CHAT, 0);
        return true;
    } 

	/* OP COMMANDS */
	if( player->userlevel < 51 ) {  ///////////////////// TRUSTED USERS AND HIGHER
		unknownCommand(player);
		return false;
    }
	
	if( strncmp( buffer, "paint ", 6 ) == 0 ) {
		player->painting = !player->painting;
		memcpy( chatbuf+2, srvMessage::ACTION_PAINT[player->painting ? 0 : 1], PACKLEN_CHAT-2 );
		send( player->sock, chatbuf, PACKLEN_CHAT, 0 );
		return true;
	}
    
    else if( strncmp( buffer, "user ", 5 ) == 0 ) {
		unsigned char level;
		Player *playerChanged;

		if( *chr == 0x20 )
			return false;

		buffer = chr;
		chr = strchr( buffer, 0x20 );
		if(chr) *chr = 0;
		else return false;

		chr++;
		/* Make sure people pass a userlevel too. */
		if( *chr == 0x20 )
			return false;

		level = (unsigned char) atoi(chr);
        memset( sprintfbuf, 0x20, sizeof(sprintfbuf) );
		if(player->userlevel < 255 && level > (player->userlevel - 50)) {
			sprintf(sprintfbuf, "&eYou cannot set user level higher than %d.", player->userlevel - 50);
            memcpy(chatbuf+2, sprintfbuf, strlen(sprintfbuf));
		    send(player->sock, chatbuf, PACKLEN_CHAT, 0);
			return false;
		}
		
        playerChanged = getPlayerByName(buffer);
        
		if(setUser(buffer, level, player) == false)
            memcpy(chatbuf+2, srvMessage::PLAYER_INFERIOR, PACKLEN_CHAT-2);
        
		else {
			sprintf( sprintfbuf, "&e%s%s%s &e(%d)", playerChanged ? playerChanged->playerName : buffer, srvMessage::ACTION_OTHERCHANGE, levelToClass(level), level );
            memcpy( chatbuf+2, sprintfbuf, PACKLEN_CHAT-2 );
        }

        send( player->sock, chatbuf, PACKLEN_CHAT, 0 );
		return true;
	}

	else if( strncmp( buffer, "tp ", 3 ) == 0 ) {
		Player *toPlayer;
        
        if( *chr == 0x20 )
			return false;

		buffer = chr;
		chr = strchr( buffer, 0x20 );
		if(chr) *chr = 0;
		else return false;

		toPlayer = getPlayerByName(buffer);

		if( ! toPlayer || (toPlayer->hidden && toPlayer->userlevel > player->userlevel) ) {
			memcpy( chatbuf+2, srvMessage::PLAYER_NOT, 19 );
		    send( player->sock, chatbuf, PACKLEN_CHAT, 0 );
			return false;
		}

		teleportPlayer( toPlayer, player );
		return true;
	}
	
    if( player->userlevel < 101 ) { ///////////////////// OPS AND HIGHER
		unknownCommand(player);
		return false;
    }
	
    if( strncmp( buffer, "k ", 2 ) == 0 || strncmp( buffer, "kick ", 5 ) == 0 ) {
		Player *toKick;
        
        if( *chr == 0x20 )
			return false;

		buffer = chr;
		chr = strchr( buffer, 0x20 );
		if(chr) *chr = 0;
		else return false;

		toKick = getPlayerByName(buffer);

		if( ! toKick ) {
			memcpy( chatbuf+2, srvMessage::PLAYER_NOT, 19 );
		    send( player->sock, chatbuf, PACKLEN_CHAT, 0 );
			return false;
		}

		if( toKick->userlevel > player->userlevel ) {
			/* Announce lack of privilege. */
			return false;
		}

		kickPlayer( toKick, srvMessage::PLAYER_KICK, true );
		return true;
	}

	else if( strncmp( buffer, "s ", 2 ) == 0 || strncmp( buffer, "say ", 4 ) == 0 ) {

		if( *chr == 0x20 )
			return false;

		memcpy( chatbuf+2, "&e", 2 );
		memcpy( chatbuf+4, chr, 66-4-j );
		broadcast( chatbuf, PACKLEN_CHAT, NULL );
		return true;
	}
	
	else if( strncmp( buffer, "hide ", 5 ) == 0 ) {
        player->hidden = !player->hidden;
        
        if( player->hidden ) {
            memcpy( chatbuf+2, srvMessage::ACTION_HIDE, 60 );
            fakeLog(player, true);
            
        } else {
            memcpy( chatbuf+2, srvMessage::ACTION_UNHIDE, 60 );
            fakeLog(player, false);
        }
            
        send( player->sock, chatbuf, PACKLEN_CHAT, 0 );

		return true;
    }

	else if( strncmp( buffer, "r ", 2 ) == 0 || strncmp( buffer, "rock ", 5 ) == 0 || strncmp( buffer, "solid ", 6 ) == 0 ) {

		if( player->block == W_BLOCK_ROCK ) {
			player->setblock = false;
			player->block = W_BLOCK_STONE;
			memcpy( chatbuf+2, srvMessage::BLOCK_NORMAL, 34 );
		}

		else {
			player->setblock = true;
			player->block = W_BLOCK_ROCK;
			memcpy( chatbuf+2, srvMessage::BLOCK_ROCK, 39 );
		}

		send( player->sock, chatbuf, PACKLEN_CHAT, 0 );
		return true;
    }

	else if( strncmp( buffer, "bring ", 6 ) == 0 ) {
		Player *who;

		if( *chr == 0x20 )
			return false;

		buffer = chr;
		chr = strchr( buffer, 0x20 );
		if(chr) *chr = 0;
		else return false;

		who = getPlayerByName(buffer);

		if( ! who ) {
			memcpy( chatbuf+2, srvMessage::PLAYER_NOT, 19 );
		    send( player->sock, chatbuf, PACKLEN_CHAT, 0 );
			return false;
		}
		
        if( who->userlevel > player->userlevel ) {
			/* Notify it can't be done, no privilege. */
			return false;
		}

		teleportPlayer( player, who );
		return true;
	}

	else if( strncmp( buffer, "draw ", 5 ) == 0 ) {
		player->drawing = true;
		player->drawTo[0] = 1;
		return true;
		//chat: place a block, baby! 
	}
    
    if( player->userlevel < 151 ) { ///////////////////// TRUSTED OPS AND HIGHER
		unknownCommand(player);
		return false;
    }

	/* Nothing yet, maybe request backups? */
		
	if( player->userlevel < 201 ) { ///////////////////// ADMINS AND HIGHER
		unknownCommand(player);
		return false;
    }

	if( strncmp( buffer, "setspawn ", 9 ) == 0 )
		return lvl->setSpawn( player->posx, player->posy, player->posz, player->rotx, player->roty );

    else if( strncmp( buffer, "b ", 2 ) == 0 || strncmp( buffer, "ban ", 4 ) == 0 ) {
		Player *toBan;

		if( *chr == 0x20 )
			return false;

		buffer = chr;
		chr = strchr( buffer, 0x20 );
		if(chr) *chr = 0;
		else return false;

		toBan = getPlayerByName(buffer);

		if(toBan) {
			if( toBan->userlevel > player->userlevel )
				return false;
		}

		banPlayer( buffer, srvMessage::PLAYER_BAN, false );
		return true;
	}

	/* Will always ban IP. */

	else if( strncmp( buffer, "unban ", 6 ) == 0 ) {

		if( *chr == 0x20 )
			return false;

		buffer = chr;
		chr = strchr( buffer, 0x20 );
		if(chr) *chr = 0;
		else return false;

		banPlayer( buffer, NULL, true );
		return true;
	}

	if(player->userlevel < 255) {
		unknownCommand(player);
		return false;
	}

	else if( strncmp( buffer, "guest ", 6 ) == 0 ) {
		player->drawing = true;
		this->guestTo[0] = 1;
		/* Chat; set block! */
		return true;
	}

	else if( strncmp( buffer, "clearguest ", 11 ) == 0 ) {
		if(this->srvGuest == false)
			return false;

		drawBlocks(0x00, guestFrom, guestTo, 0);
		return true;
	}
	
	return false;
}

void Server::releasePlayer(Player *player) {
	unsigned int i;
	char buffer[128];
	for(i = 0; i < playerList.size(); i++) {
		if(&playerList.at(i) == player) {
			memset(buffer, 0, sizeof(buffer));
			sprintf(buffer, "%s (%s) has disconnected.\n", playerList.at(i).playerName, inet_ntoa(playerList.at(i).clientinfo.sin_addr));
			Logger::add(buffer);

			buffer[0] = 0x0C;
			buffer[1] = playerList.at(i).id_byte;

			if(playerList.at(i).hidden == true)
				broadcastLevelSelective(buffer, 2, &playerList.at(i), false);
			else
				broadcast(buffer, 2, NULL);

			this->freeIdList[playerList.at(i).id_byte] = 0;
			FD_CLR(playerList.at(i).sock, &master);
			closesocket(playerList.at(i).sock);
			playerList.erase(playerList.begin() + i);
		}
	}

	this->isOperated = isOperatorOn();
	this->playerNum = playerList.size();
}

void Server::kickPlayer( Player *player, const char *kickMsg, bool announce ) {
	char kickbuf[PACKLEN_CHAT];

	memset(kickbuf, 0x20, PACKLEN_KICK);

	kickbuf[0] = 0x0E;
	memcpy(kickbuf+1, kickMsg, strlen(kickMsg));
	send(player->sock, kickbuf, PACKLEN_KICK, 0);

	if(announce) {
		memset(kickbuf, 0x20, PACKLEN_CHAT);
		kickbuf[0] = 0x0D;
		kickbuf[1] = 0x00;
		memcpy(kickbuf+2, "&e", 2);
		memcpy(kickbuf+4, player->playerName, player->namelen);
		memcpy(kickbuf+4+player->namelen, srvMessage::ACTION_KICK, 28);
		broadcast(kickbuf, PACKLEN_CHAT, NULL);
	}

	releasePlayer(player);
}

void Server::banPlayer( char *playerName, const char *banMsg, bool unban ) {
	char banbuf[PACKLEN_CHAT], playerCompare[32];
	char ipbuf[16];
	struct Banned *tmp;
	Player *toBan;
	FILE *banned;

	toBan = getPlayerByName(playerName);

	if(toBan) strcpy(ipbuf, inet_ntoa(toBan->clientinfo.sin_addr));
	else ipbuf[0] = 0;

	/* Do not ban twice. */
	if( unban == false && isBanned(playerName, ipbuf) )
		return;

	/* Do not unban someone who's not banned :P */
	else if( unban == true && ! isBanned(playerName, ipbuf) )
		return;

	strcpy(playerCompare, playerName);
	toLowerCase(playerCompare);

	if(toBan) {
		kickPlayer(toBan, banMsg, false);
		memset(banbuf, 0x20, PACKLEN_CHAT);
		banbuf[0] = 0x0D;
		banbuf[1] = 0x00;
		memcpy(banbuf+2, "&e", 2);
		memcpy(banbuf+4, toBan->playerName, toBan->namelen);
		memcpy(banbuf+4+toBan->namelen, srvMessage::ACTION_BAN, 12);
		broadcast(banbuf, PACKLEN_CHAT, NULL);
	}

	tmp = this->bannedList;

	while(tmp->next) {
		if(unban) {
			if( strcmp( tmp->name, playerCompare ) == 0 ) {
				tmp->name[0] = 0;
				tmp->ip[0] = 0;
			}
		}
		tmp = tmp->next;
	}

	/* This is a little hackish, but it's faster then iterating -
	   over the list twice. */
	if(unban) {
		if( strcmp( tmp->name, playerCompare ) == 0 ) {
			tmp->name[0] = 0;
			tmp->ip[0] = 0;
		}
	}

	else {
		tmp->next = (struct Banned*) malloc( sizeof(struct Banned) );
		tmp = tmp->next;
		memset( tmp, 0, sizeof(struct Banned) );
		strcpy( tmp->name, playerCompare );
		strcpy( tmp->ip, ipbuf );
	}

	tmp = this->bannedList;
	banned = fopen( "banned.txt", "w" );

	while(tmp) {
		if( tmp->name[0] || tmp->ip[0] )
			fprintf( banned, "%s:%s\n", tmp->name, tmp->ip );
		tmp = tmp->next;
	}

	fclose(banned);
}

User *Server::getUserByName( char *playerName ) {
	struct User *tmp;
	tmp = this->userList;
    while(tmp) {
		if( strcmp( tmp->name, playerName ) == 0 ) {		/* Ugly but faster. */
			return tmp;
		}
		tmp = tmp->next;
	}
	return NULL;
}


void Server::refreshColor( Player *player ) {
	char logbuf[PACKLEN_SPAWN];
		
	logbuf[0] = 0x0C;
	logbuf[1] = player->id_byte;
	broadcast(logbuf, 2, player);
        
	memset(logbuf, 0x20, PACKLEN_SPAWN);

	logbuf[0] = 0x07;
	logbuf[1] = player->id_byte;

	logbuf[2] = '&';
	logbuf[3] = getColorCode(player->userlevel);
	memcpy(logbuf+4, player->playerName, player->namelen);
	memcpy(logbuf+66, &player->posx, 2);
	memcpy(logbuf+68, &player->posy, 2);
	memcpy(logbuf+70, &player->posz, 2);
	logbuf[72] = player->rotx;
	logbuf[73] = player->roty;
	broadcast(logbuf, PACKLEN_SPAWN, player);
}

bool Server::setUser( char *playerName, unsigned char level, Player *changer ) {
	char playerCompare[32];
	char notifybuf[PACKLEN_CHAT];
	User *user;
	Player *player;
	FILE *users;

	strcpy(playerCompare, playerName);
	toLowerCase(playerCompare);

	user = getUserByName(playerCompare);
	if(user) {
		if( changer->userlevel != 255 && changer->userlevel <= user->level ) {
			/* Notify changer this can't be done. */
			return false;
		}
		user->level = level;
	}

	else {
		user = this->userList;
		while(user->next) user = user->next;
		user->next = (User*) malloc( sizeof(struct User) );
		user = user->next;
		strcpy(user->name, playerCompare);
		user->level = level;
		user->next = NULL;
	}

	player = getPlayerByName(playerCompare);
	if(player) {
		player->userlevel = level;
		player->isoperator = (player->userlevel > 100);
		memset(notifybuf, 0x20, PACKLEN_CHAT);
		notifybuf[0] = 0x0D;
		notifybuf[1] = 0x00;
		memcpy(notifybuf+2, srvMessage::ACTION_LEVELCHANGE, 64);
		sprintf(notifybuf+17, "%s &e(%d)", levelToClass(player->userlevel), player->userlevel);
		send(player->sock, notifybuf, PACKLEN_CHAT, 0);
		refreshColor(player);
	}

	user = this->userList;
	users = fopen( "users.txt", "w" );

	while(user) {
		if( user->name[0] )
			fprintf( users, "%s:%d\n", user->name, user->level );
		user = user->next;
	}

	fclose(users);
	return true;
}

bool Server::verifyName( Player *player, char *hash ) {
	unsigned char sig[16];
	char hashed[33], hashbuf[64];

	hashbuf[0] = 0;		/* No need to memset. */
	sprintf(hashbuf, "%d%s", this->salt, player->playerName);
	md5_buffer(hashbuf, strlen(hashbuf), sig);
	md5_sig_to_string(sig, hashed, sizeof(hashed));

	if(strcmp(hash, hashed) == 0)
		return true;

	return false;
}

void Server::beat() {
	unsigned int sock;
	char httpheader[300], hbeat[256];
	static bool firstBeat = true;

	memset(hbeat, 0, sizeof(hbeat));
	memset(httpheader, 0, sizeof(httpheader));
	sprintf(hbeat, "port=%s&users=%d&max=%d&name=%s&public=%s&version=7&salt=%d", this->srvPort, this->playerNum, this->playerMax, this->srvName, this->srvPublic ? "true" : "false", this->salt);
	sprintf(httpheader, "POST /heartbeat.jsp HTTP/1.1\r\nContent-Type:application/x-www-form-urlencoded\r\nContent-Length:%d\r\nContent-Language: en-US\r\nCache-Control: no-cache\r\nPragma:no-cache\r\nUser-Agent: minerCPP\r\nHost:www.minecraft.net\r\nAccept: text/html, image/gif, image/jpeg, *; q=.2,*/*; q=.2\r\nConnection: keep-alive\r\n\r\n", strlen(hbeat));

    sock = socket(heartbeat->ai_family, heartbeat->ai_socktype, heartbeat->ai_protocol);
    connect(sock, heartbeat->ai_addr, heartbeat->ai_addrlen);
    send(sock, httpheader, strlen(httpheader), 0);
    send(sock, hbeat, strlen(hbeat), 0);

	if(firstBeat == true) {
		char hbeatrecv[300];
		char *beatptr, *msgptr;
		FILE *exurl;

		recv(sock, hbeatrecv, sizeof(hbeatrecv), 0);
		beatptr = strstr(hbeatrecv, "http://");

		if(beatptr != NULL) {
			exurl = fopen("external_url.txt", "w");
			fwrite(beatptr, 1, 73, exurl);
			fclose(exurl);
			msgptr = "Ext. URL saved to file: external_url.txt\n";
			firstBeat = false;
		}

		else
			msgptr = "Couldn't retrieve ext. URL. Will try again on next heartbeat.\n";

		Logger::add(msgptr);
	}

	closesocket(sock);
}

bool Server::isOperatorOn() {
	unsigned int i;

	for( i = 0; i < playerList.size(); i++ ) {
		if(playerList.at(i).isoperator == true)
			return true;
	}

	return false;
}

/* Exclude can be used for sending stuff to everyone but "this one player". */
void Server::broadcast( char *packet, int paklen, Player *exclude ) {
	unsigned int i;
	for( i = 0; i < playerList.size(); i++ ) {
		if( &playerList.at(i) == exclude )
			continue;

		send( playerList.at(i).sock, packet, paklen, 0 );
	}
}

bool Server::privsay(Player *player, char *buffer) {
	int i = 63;
	char chatbuf[PACKLEN_CHAT], *chr;
	Player *toPlayer;

	if(*buffer == 0x20) {
		buffer++;
		if(*buffer == 0x20) {
			/* Player has requested to flush his id_privsay... */
			player->id_privsay = player->id_byte;
			return false;
		}
		i--;
	}
	
	chr = (char*) memchr(buffer, 0x20, i);
	if(chr) *chr = 0x00;
	else return false;

	if( (chr - buffer) > 32 )
		return 0;

	toPlayer = getPlayerByName(buffer);
	
	if(toPlayer == NULL) {
		/* No such player, notify. */
		return false;
	}

	player->id_privsay = toPlayer->id_byte;	/* This should allow higher ranked, hidden players to initialize a char with lowers. */

	if(toPlayer == player || (toPlayer->hidden && player->userlevel < toPlayer->userlevel && player->id_byte != toPlayer->id_privsay)) {
		/* Do not pm yourself, faggot. */
		/* Nor somebody of a higher userlevel, that is currently hidden :D */
		return false;
	}

	chr++;

	if(*chr == 0x20)		/* Disallow blanks for now. Temporary. */
		return false;

	memset( chatbuf, 0x20, PACKLEN_CHAT );

	chatbuf[0] = 0x0D;
	chatbuf[1] = 0x00;
	chatbuf[2] = 0x00; /* to avoid using sprintfbuf */
	sprintf(chatbuf+2, "&6(&%c%s&6) >>&7 ", getColorCode(toPlayer->userlevel), toPlayer->playerName);
	i = strlen(chatbuf+2);
	memcpy(chatbuf+2+i, chr, PACKLEN_CHAT-i-3);
	send(player->sock, chatbuf, PACKLEN_CHAT, 0);

	chatbuf[2] = 0x00; /* to avoid using sprintfbuf */
	sprintf(chatbuf+2, "&6(&%c%s&6):&7 ", getColorCode(player->userlevel), player->playerName);
	i = strlen(chatbuf+2);
	memcpy(chatbuf+2+i, chr, PACKLEN_CHAT-i-3);
	send(toPlayer->sock, chatbuf, PACKLEN_CHAT, 0);

	return true;
}

void Server::broadcastLevelSelective( char *packet, int paklen, Player *who, bool includeLower ) {
	unsigned int i;
	for( i = 0; i < playerList.size(); i++ ) {
		if( (&playerList.at(i) == who) || (! includeLower == playerList.at(i).userlevel < who->userlevel) )
			continue;

		send( playerList.at(i).sock, packet, paklen, 0 );
	}
}

char Server::getColorCode(unsigned char userlevel) {
	if(userlevel > 250)
		return '8';
	else if(userlevel > 200)
		return 'c';
	else if(userlevel > 100)
		return '9';
	else if(userlevel > 50)
		return 'a';
	else if(userlevel > 0)
		return 'f';
	return '7';
}

/* Parses packets sent by the clients */
/* and performs all needed actions */
bool Server::parsePacket( Player *player, char *buffer ) {
	int i;
	char buf[128], *chr;

    /* Login */
	if( *buffer == 0x00 && *(buffer+1) == 0x07 ) {
		buffer += 2;
		chr = strchr(buffer, 0x20);

		if(chr) *chr = 0;
		else return false;

		strcpy(player->playerName, buffer);
		player->namelen = strlen(player->playerName);
		memset(buf, 0, sizeof(buf));

		if( isBanned(player->playerName, inet_ntoa(player->clientinfo.sin_addr)) ) {
			kickPlayer( player, "You are banned!", false );
			sprintf( buf, "%s (%s) tried to connect, but is banned.\n", player->playerName, inet_ntoa(player->clientinfo.sin_addr) );
			Logger::add(buf);
			return false;
		}

		if( this->srvVerify && this->srvPublic ) {
			buffer += 64;
			*(buffer+32) = 0;

			if( verifyName(player, buffer) == false ) {
				kickPlayer( player, "Illegal name.", false );
				sprintf( buf, "%s (%s) tried to connect, but could not verify his name.\n", player->playerName, inet_ntoa(player->clientinfo.sin_addr) );
				Logger::add(buf);
				return false;
			}
		}

		sprintf( buf, "%s connected from IP: %s\n", player->playerName, inet_ntoa(player->clientinfo.sin_addr) );
		Logger::add(buf);

		if( isListed(player) ) {
            player->isoperator = (player->userlevel > 100);
			this->isOperated = isOperatorOn();
		}

		sendServerInfo(player);
		sendMap(player);
		sendPlayerList(player);

		/* WELCOME MESSAGES HERE */
		memset( buf, 0x20, PACKLEN_CHAT );
		buf[0] = 0x0D;
		buf[1] = 0x00;
		memcpy( buf+2, "&9Type &7/help &9for a list of commands!", 40 );
		send( player->sock, buf, PACKLEN_CHAT, 0 );
		return true;
	}

	/* Movement */
	else if( *buffer == 0x08 ) {
        *(buffer+1) = player->id_byte;
            
        memcpy( &player->posx, buffer+2, 2 );
        memcpy( &player->posy, buffer+4, 2 );
        memcpy( &player->posz, buffer+6, 2 );
        player->rotx = *(buffer+8);
        player->roty = *(buffer+9);

		if( ! player->hidden )
            broadcast( buffer, 10, player );
        else
            broadcastLevelSelective( buffer, 10, player, false );
            
        return true;
	}

	/* Blockchange */
	else if( *buffer == 0x05 ) {
		unsigned short x, y, z;
		unsigned char block, changedblock;
		bool ricochet = false;

		memcpy( &x, buffer+1, 2 );
		memcpy( &y, buffer+3, 2 );
		memcpy( &z, buffer+5, 2 );
		x = ntohs(x);
		y = ntohs(y);
		z = ntohs(z);

		*buffer = 0x06;
		block = *(buffer+8);

		if(block > 0x31 || (lvl->validBlock(x, y, z) == false))
            return false;

		/* Stupid idea then, stupid idea now. :| */
		/* noteBlockChange(x, y, z, player); */
		
		memset(buf, 0x20, PACKLEN_CHAT);
		buf[0] = 0x0D;
		buf[1] = 0x00;

		if(*(buffer+7) || player->painting == true) {
			if(player->setblock == true && block == W_BLOCK_STONE)
					block = player->block;
			ricochet = true;
		}

		else block = W_BLOCK_AIR;

		/* We'll do it this way until we thought of something better! */
		if(player->userlevel > 100 && player->drawing == true) {

			char chatbuf[PACKLEN_CHAT];

			if(this->guestTo[0] == 1 && player->userlevel == 255 ) {
				this->guestFrom[0] = x;
				this->guestFrom[1] = y;
				this->guestFrom[2] = z;
				this->guestTo[0] = 9999;
			}

			else if(this->guestTo[0] == 9999 && player->userlevel == 255 ) {
				this->guestTo[0] = x;
				this->guestTo[1] = y;
				this->guestTo[2] = z;

				setGuestArea();
				memset(chatbuf, 0x20, PACKLEN_CHAT);
				chatbuf[0] = 0x0D;
				chatbuf[1] = 0x00;
				memcpy(chatbuf+2, "&eGuestarea recorded!", 21);
				send(player->sock, chatbuf, PACKLEN_CHAT, 0);

				player->drawing = false;
			}

			else if(player->drawTo[0] == 1) {
				player->drawFrom[0] = x;
				player->drawFrom[1] = y;
				player->drawFrom[2] = z;
				player->drawTo[0] = 9999;
			}
			
			else if(player->drawTo[0] == 9999) {
				player->drawTo[0] = x;
				player->drawTo[1] = y;
				player->drawTo[2] = z;
				drawBlocks(block, player->drawFrom, player->drawTo, player->userlevel > 255 ? 0 : player->userlevel);
				player->drawing = false;
			}

		}

		changedblock = lvl->getBlock(x, y, z);

		if( ((block == W_BLOCK_ROCK || changedblock == W_BLOCK_ROCK) && player->isoperator == false) ||
			(player->userlevel < 1 && isGuestBlock(x, y, z) == false) ||
			(player->userlevel < 51 && isOperatorOn() == false && isGuestBlock(x, y, z) == false) ) {
				*(buffer+7) = changedblock;
				send(player->sock, buffer, 8, 0);
				return false;
		}

		if(block == W_BLOCK_STAIR && lvl->validBlock(x, y-1, z) && lvl->getBlock(x, y-1, z) == W_BLOCK_STAIR) {
			*(buffer+7) = W_BLOCK_AIR;
			send( player->sock, buffer, 8, 0 );
			y--;
			y = htons(y);
			memcpy( buffer+3, &y, 2 );
			y = ntohs(y);
			broadcast( buffer, 8, NULL );
			block = W_BLOCK_DOUBSTAIR;
			ricochet = true;
		}

		*(buffer+7) = block;
		lvl->setBlock(x, y, z, block);
		broadcast( buffer, 8, ricochet ? NULL : player );
		return true;
	}

	/* Chat */
	else if( *buffer == 0x0D ) {

		for( i = 2; i < PACKLEN_CHAT; ++i )
			if( *(buffer+i) < 0x20 || *(buffer+i) > 0x7E || *(buffer+i) == '&' )
				*(buffer+i) = 0x01; 
		
		if( *(buffer+2) == '/' ) {
			buffer += 3;
			Logger::add_adminEntry( buffer, player );
			return playerCommand( player, buffer );
		}

		else if( *(buffer+2) == '@' ) {
			buffer += 3;
			return privsay(player, buffer);
		}

		buffer += 2;

		Logger::add_chatEntry(buffer, player);

		memset(buf, 0x20, PACKLEN_CHAT);

		buf[0] = 0x0D;
		buf[1] = 0x00;
		buf[2] = '&';
		buf[3] = getColorCode(player->userlevel);

		i = strlen(player->playerName);
		memcpy(buf+4, player->playerName, player->namelen);
		memcpy(buf+player->namelen+4, "&f: ", 4);
		memcpy(buf+i+8, buffer, PACKLEN_CHAT-i-8);
		broadcast(buf, PACKLEN_CHAT, NULL);
		return true;
	}

	return false;
}

void Server::sendServerInfo( Player *player ) {
	static char infoBuffer[PACKLEN_SRVINFO];
	static bool doneThis = false;

	/* Space out these strings ONCE only. */
	if(doneThis == false) {
		unsigned int i;

		memset(infoBuffer, 0x20, sizeof(infoBuffer));
		infoBuffer[0] = 0x00;
		infoBuffer[1] = 0x07;

		i = strlen(this->srvName);
		memcpy(infoBuffer+2, this->srvName, i);

		i = strlen(this->srvMotd);
		memcpy(infoBuffer+2+64, this->srvMotd, i);

		doneThis = true;
	}

	/* > 50 = > TRUSTEDS ONLY */
	infoBuffer[2+64+64] = (player->userlevel > 100 ? 0x64 : 0x00);

	send(player->sock, infoBuffer, PACKLEN_SRVINFO, 0);
}

void Server::sendPlayerList( Player *player ) {
	unsigned int i;
	char spawnbuf[PACKLEN_SPAWN], colorCode;

	for( i = 0; i < playerList.size(); i++ ) {

		memset( spawnbuf, 0x20, PACKLEN_SPAWN );
		spawnbuf[0] = 0x07;

		colorCode = getColorCode(playerList.at(i).userlevel);

		if( playerList.at(i).id_byte == player->id_byte ) {
			/* Send spawn to player who joined. */
            
			spawnbuf[1] = -1;

			spawnbuf[2] = '&';
			spawnbuf[3] = colorCode;
			memcpy(spawnbuf+4, player->playerName, player->namelen);

			memcpy(spawnbuf+66, &player->posx, 2);
			memcpy(spawnbuf+68, &player->posy, 2);
			memcpy(spawnbuf+70, &player->posz, 2);
			spawnbuf[72] = player->rotx;
			spawnbuf[73] = player->roty;

			send(player->sock, spawnbuf, PACKLEN_SPAWN, 0);

			spawnbuf[1] = player->id_byte;
			broadcast(spawnbuf, PACKLEN_SPAWN, player);

			spawnbuf[0] = 0x0D;
			spawnbuf[1] = 0x00;
			memcpy(spawnbuf+2, "&e", 2);
			memcpy(spawnbuf+4, player->playerName, player->namelen);
			memcpy(spawnbuf+player->namelen+4, " joined the game!", 17);
			broadcast(spawnbuf, PACKLEN_CHAT, NULL);
		}

		else {
			/* Send player at playerList[i] to player. */
		    if(playerList.at(i).hidden && (playerList.at(i).userlevel > player->userlevel))
			     continue; /* Don't show hidden players */
			spawnbuf[1] = playerList.at(i).id_byte;
			spawnbuf[2] = '&';
			spawnbuf[3] = colorCode;
			memcpy(spawnbuf+4, playerList.at(i).playerName, playerList.at(i).namelen);
			memcpy(spawnbuf+66, &playerList.at(i).posx, 2);
			memcpy(spawnbuf+68, &playerList.at(i).posy, 2);
			memcpy(spawnbuf+70, &playerList.at(i).posz, 2);
			spawnbuf[72] = playerList.at(i).rotx;
			spawnbuf[73] = playerList.at(i).roty;
			send( player->sock, spawnbuf, PACKLEN_SPAWN, 0 );
		}

	}
}

void Server::sendMap( Player *player ) {
	unsigned long sent, left, total;
	unsigned short chunksize;
	unsigned char mapbuf[PACKLEN_MAP], *map;
	unsigned char percentage;

	map = (unsigned char *) malloc( lvl->getCompressedMapLen() + 16384 );
	lvl->setCompressedMapLen( gzip_compress( map, lvl->getServerLevel() + OUR_HEADERLEN, lvl->getDecompressedMapLen() - OUR_HEADERLEN ) );

	send( player->sock, "\x02", 1, 0 );

	sent = 0;
	left = lvl->getCompressedMapLen();
	total = left;

	while(sent < total) {
		mapbuf[0] = 0x03;

		left = total - sent;
		chunksize = (left < 1024 ? (unsigned short) left : 1024);
		sent += chunksize;

		memcpy(mapbuf+3, map, chunksize);
		map += chunksize;

		percentage = (unsigned char) ((sent * 100) / total);
		mapbuf[3+chunksize] = percentage;

		/* Weird doing this last but it reduce one call to htons. */
		chunksize = htons(chunksize);
		memcpy(mapbuf+1, &chunksize, 2);

		send(player->sock, (const char*) mapbuf, PACKLEN_MAP, 0);

		/* How's this for not threading? ;D */
		this->step();
	}

	mapbuf[0] = 0x04;

	chunksize = lvl->getDimX();
	memcpy(mapbuf+1, &chunksize, 2);

	chunksize = lvl->getDimY();
	memcpy(mapbuf+3, &chunksize, 2);

	chunksize = lvl->getDimZ();
	memcpy(mapbuf+5, &chunksize, 2);

	send(player->sock, (const char*) mapbuf, 7, 0);

	map -= sent;
	free(map);
}

bool Server::isBanned( char *playerName, char *ip ) {
	struct Banned *tmp;
	char playerCompare[32];

	strcpy(playerCompare, playerName);
    toLowerCase(playerCompare);
    
    tmp = this->bannedList;

	while(tmp) {
		if( strcmp( tmp->name, playerCompare ) == 0 ) 
			return true;
		else if( ip[0] && strcmp( tmp->ip, ip ) == 0 )
			return true;
		tmp = tmp->next;
	}

	return false;
}


bool Server::isListed( Player *player ) {
	struct User *tmp;
	char playerCompare[32];

	strcpy(playerCompare, player->playerName);
    toLowerCase(playerCompare);

	tmp = this->userList;

	while(tmp) {
		if( strcmp( tmp->name, playerCompare ) == 0 ) {
				player->userlevel = (tmp->level > 0 ? tmp->level : 1);
				return true;
		}
		tmp = tmp->next;
	}

	return false;
}

Player *Server::getPlayerByName(char *playerName) {
    unsigned int i;
	char playerCompare1[32], playerCompare2[32];

	if(strlen(playerName) > 32)
		return 0;

	strcpy(playerCompare1, playerName);
	toLowerCase(playerCompare1);

    for( i = 0; i < playerList.size(); i++ ) {
        strcpy( playerCompare2, playerList.at(i).playerName );
		toLowerCase(playerCompare2);
                
		if( strcmp( playerCompare1, playerCompare2 ) == 0 )
			return &playerList.at(i);
	}

	return NULL;
}

Player *Server::getPlayerById( unsigned char id_byte ) {
	unsigned int i;

	for( i = 0; i < playerList.size(); i++ ) {
		if( playerList.at(i).id_byte == id_byte )
			return &playerList.at(i);
	}

	return NULL;
}

unsigned char Server::getFreeId() {
	unsigned char iterator;
	for( iterator = 0; this->freeIdList[iterator] == 0xFF; iterator++ );
	this->freeIdList[iterator] = 0xFF;
	return iterator;
}
