#ifndef _SERVER_H
#define _SERVER_H

class Server {

public:

	Server();
	~Server();

	void init();
	bool step();

private:

	void beat();
	void broadcast( char *packet, int paklen, Player *exclude );
	void broadcastLevelSelective( char *packet, int paklen, Player *who, bool includeLower );

	bool parsePacket( Player *player, char *buffer );
	bool privsay(Player *player, char *buffer);
	void sendServerInfo( Player *player );
	void sendPlayerList( Player *player );
	void sendMap( Player *player );

	bool verifyName( Player *player, char *hash );
	bool playerCommand( Player *player, char *buffer );
	void unknownCommand( Player *player );
	void fixChars( char *buffer );
	void releasePlayer(Player *player);

	void teleportPlayer( Player *toPlayer, Player *who );
	void kickPlayer( Player *player, const char *kickMsg, bool announce );
	void banPlayer( char *playerName, const char *banMsg, bool unban );
	void fakeLog(Player *player,  bool out);
	void refreshColor(Player *player);
	bool drawBlocks(char block, unsigned short *drawFrom, unsigned short *drawTo, unsigned char drawMax);

	bool isListed( Player *player );
	bool isBanned( char *playerName, char *ip );
	bool setUser( char *playerName, unsigned char level, Player *changer );
	bool isOperatorOn();
	void setGuestArea();
	bool isGuestBlock(unsigned short x, unsigned short y, unsigned short z);

	User *getUserByName( char *playerName );
	Player *getPlayerByName( char *playerName );
	Player *getPlayerById(unsigned char id_byte);
	char *levelToClass(unsigned char userlevel);
	char getColorCode(unsigned char userlevel);

	unsigned char getFreeId();

	int playerNum;
	int playerMax;

	int backupInterval;

	char srvName[64];
	char srvMotd[64];
	char srvPort[8];
	bool srvPublic;
	bool srvVerify;

	unsigned char freeIdList[UCHAR_MAX];
	
	unsigned long ticks, ticks_beat, ticks_ping, ticks_savemap, ticks_backupmap;
	unsigned long salt;

	bool isOperated;
	
	bool srvGuest;
	unsigned short guestFrom[3];
	unsigned short guestTo[3];

	WSADATA wsaData;
	fd_set master, readfds;
	struct timeval timeout;
	unsigned int sock_listener;
	struct addrinfo *heartbeat, *listener;

	vector<Player> playerList;
	struct User *userList;
	struct Banned *bannedList;

	Level *lvl;
	Config *cfg;

};

#endif
